@@ -1,6 +1,6 @@
 //===========================================================================
 // @(#) $Name:$
-// @(#) $Id: DwmAuthPeerAuthenticator.cc 10617 2020-05-02 21:18:04Z dwm $
+// @(#) $Id: DwmAuthPeerAuthenticator.cc 10934 2020-08-17 04:03:47Z dwm $
 //===========================================================================
 //  Copyright (c) Daniel W. McRobb 2017
 //  All rights reserved.
@@ -63,7 +63,7 @@ extern "C" {
 #include "DwmAuthPeerAuthenticator.hh"
 #include "DwmAuthSymCrypto.hh"
 
-static const Dwm::SvnTag svntag("@(#) $DwmPath: dwm/libDwmAuth/tags/libDwmAuth-0.3.3/src/DwmAuthPeerAuthenticator.cc 10617 $");
+static const Dwm::SvnTag svntag("@(#) $DwmPath: dwm/libDwmAuth/tags/libDwmAuth-0.3.4/src/DwmAuthPeerAuthenticator.cc 10934 $");
 
 using namespace std;
 using CryptoPP::PSS;
@@ -172,6 +172,99 @@ namespace Dwm {
       return rc;
     }
 
+    //------------------------------------------------------------------------
+    //!  
+    //------------------------------------------------------------------------
+    static bool WriteString(boost::asio::ip::tcp::socket & sck,
+                            const std::string & str)
+    {
+      using boost::asio::write, boost::asio::buffer,
+        boost::system::error_code;
+
+      bool        rc = false;
+      uint32_t    len = str.size();
+      uint32_t    nwlen = htonl(len);
+      error_code  ec;
+      if ((write(sck, buffer(&nwlen, sizeof(nwlen)), ec) == sizeof(nwlen))
+          && (! ec)) {
+        if (len) {
+          if ((write(sck, buffer(str.data(), len), ec) == len) && (! ec)) {
+            rc = true;
+          }
+        }
+        else {
+          rc = true;
+        }
+      }
+      return rc;
+    }
+
+    //------------------------------------------------------------------------
+    //!  
+    //------------------------------------------------------------------------
+    static bool ReadString(boost::asio::ip::tcp::socket & sck,
+                           std::string & str)
+    {
+      using boost::asio::read, boost::system::error_code, boost::asio::buffer;
+      
+      str.clear();
+      bool        rc = false;
+      uint32_t    nwlen;
+      error_code  ec;
+      if ((read(sck, buffer(&nwlen, sizeof(nwlen)), ec) == sizeof(nwlen))
+          && (! ec)) {
+        uint32_t  len = htonl(nwlen);
+        if (len) {
+          try {
+            str.resize(len);
+            if ((read(sck, buffer(str.data(), str.size()), ec)
+                 == str.size())
+                && (! ec)) {
+              rc = true;
+            }
+          }
+          catch (...) {
+            Syslog(LOG_ERR, "Exception resizing string to %u characters",
+                   len);
+            str.clear();
+          }
+        }
+        else {
+          rc = true;
+        }
+      }
+      return rc;
+    }
+    
+    //------------------------------------------------------------------------
+    //!  
+    //------------------------------------------------------------------------
+    bool PeerAuthenticator::ECExchange(boost::asio::ip::tcp::socket & s,
+                                       Auth::ECDHAgreement & agreement)
+    {
+      bool  rc = false;
+      string  mypubEC = to_str(agreement.Public());
+      if (WriteString(s, mypubEC)) {
+        string  theirPubECDH;
+        if (ReadString(s, theirPubECDH)) {
+          agreement.TheirPublic(to_SecByteBlock(theirPubECDH));
+          if (agreement.Agree()) {
+            rc = true;
+          }
+          else {
+            Syslog(LOG_ERR, "Failed to reach agreement");
+          }
+        }
+        else {
+          Syslog(LOG_ERR, "Read() failed");
+        }
+      }
+      else {
+        Syslog(LOG_ERR, "Write() failed");
+      }
+      return rc;
+    }
+    
     //------------------------------------------------------------------------
     //!  
     //------------------------------------------------------------------------
@@ -241,6 +334,47 @@ namespace Dwm {
       return rc;
     }
 
+    //------------------------------------------------------------------------
+    //!  
+    //------------------------------------------------------------------------
+    bool PeerAuthenticator
+    ::IDAndSigExchange(boost::asio::ip::tcp::socket & sck,
+                       const ECDHAgreement & agreement,
+                       string & theirId,
+                       string & theirIdCrypted,
+                       string & theirSig)
+    {
+      bool  rc = false;
+      string idAndSig = BuildIDAndSignature(agreement);
+      if (WriteString(sck, idAndSig)) {
+        string  s;
+        if (ReadString(sck, s)) {
+          string  keystr((to_str(agreement.AgreedKey())).substr(0, 16));
+          string  ivstr((to_str(agreement.AgreedKey())).substr(16, 16));
+          istringstream  is(s);
+          if (IO::Read(is, theirIdCrypted)) {
+            theirId = SymCrypto::Decrypt(keystr, ivstr, theirIdCrypted);
+            if (IO::Read(is, theirSig)) {
+              rc = true;
+            }
+            else {
+              Syslog(LOG_ERR, "Failed to read signature");
+            }
+          }
+          else {
+            Syslog(LOG_ERR, "Failed to read encrypted ID");
+          }
+        }
+        else {
+          Syslog(LOG_ERR, "Failed to read ID and signature");
+        }
+      }
+      else {
+        Syslog(LOG_ERR, "Failed to write ID and signature");
+      }
+      return rc;
+    }
+    
     //------------------------------------------------------------------------
     //!  
     //------------------------------------------------------------------------
@@ -301,7 +435,57 @@ namespace Dwm {
       }
       return rc;
     }
-    
+
+    //------------------------------------------------------------------------
+    //!  
+    //------------------------------------------------------------------------
+    bool PeerAuthenticator::
+    Authenticate(boost::asio::ip::tcp::socket & sck, string & theirId,
+                 string & agreedKey)
+    {
+      bool  rc = false;
+      theirId.clear();
+      agreedKey.clear();
+      Auth::ECDHAgreement  agreement;
+      if (ECExchange(sck, agreement)) {
+        string  theirIdCrypted, theirSig;
+        if (IDAndSigExchange(sck, agreement, theirId,
+                             theirIdCrypted, theirSig)) {
+          Auth::PublicKeysFile  pubKeysFile;
+          pubKeysFile.Load(_pubKeysPath);
+          Ed25519::PublicKey  theirPubKey;
+          if (pubKeysFile.Find(theirId, theirPubKey)) {
+            string  msgToVerify =
+              to_str(agreement.TheirPublic()) + theirIdCrypted;
+            CryptoPP::ByteQueue  bt;
+            theirPubKey.Save(bt);
+            CryptoPP::ed25519::Verifier  verifier(bt);
+            if (verifier.VerifyMessage((const uint8_t *)msgToVerify.c_str(),
+                                       msgToVerify.size(),
+                                       (const uint8_t *)theirSig.c_str(),
+                                       theirSig.size())) {
+              agreedKey = to_str(agreement.AgreedKey());
+              rc = true;
+            }
+            else {
+              Syslog(LOG_ERR, "Failed to verify message from %s",
+                     theirId.c_str());
+            }
+          }
+          else {
+            Syslog(LOG_ERR, "Failed to find %s in %s",
+                   theirId.c_str(), _pubKeysPath.c_str());
+          }
+        }
+        else {
+          Syslog(LOG_ERR, "ID and signature exhange failed");
+        }
+      }
+      else {
+        Syslog(LOG_ERR, "ECDH exhange failed");
+      }
+      return rc;
+    }
 
   }  // namespace Auth
 
